#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <GL/glew.h>
#ifndef __APPLE__
#include <GL/glut.h>
#else
#include <GLUT/glut.h>
#endif
#include "sdr.h"
#include "teapot_data.h"

void disp(void);
void draw_teapot(float scale);
void draw_teapot_patch(struct vec3 *bez_cp, int *index, int flip, float rot);
void set_material(float dr, float dg, float db, float sr, float sg, float sb, float shin);
void reshape(int x, int y);
void keyb(unsigned char key, int x, int y);
void mouse(int bn, int state, int x, int y);
void motion(int x, int y);

static float cam_theta, cam_phi = 25;
static float cam_dist = 8.0;

static unsigned int prog;

static int max_tess_level;
static int tess_level = 10;

int main(int argc, char **argv)
{
	float amb[] = {0.05, 0.05, 0.05, 0.0};

	glutInit(&argc, argv);
	glutInitWindowSize(1280, 800);
	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
	glutCreateWindow("tesselation shaders test");

	glutDisplayFunc(disp);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyb);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);

	glewInit();

	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, amb);
	glClearColor(amb[0], amb[1], amb[2], 1);

	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);

	if(!GLEW_ARB_tessellation_shader) {
		fprintf(stderr, "your OpenGL implementation does not support tesselation shaders\n");
		return 1;
	}
	glGetIntegerv(GL_MAX_TESS_GEN_LEVEL, &max_tess_level);
	printf("maximum tesselation levels: %d\n", max_tess_level);

	glPatchParameteri(GL_PATCH_VERTICES, 16);

	{
		unsigned int vsdr, tcsdr, tesdr, psdr;

		if(!(vsdr = load_vertex_shader("sdr/bezier.v.glsl"))) {
			return 1;
		}
		if(!(tcsdr = load_tessctl_shader("sdr/bezier.tc.glsl"))) {
			return 1;
		}
		if(!(tesdr = load_tesseval_shader("sdr/bezier.te.glsl"))) {
			return 1;
		}
		if(!(psdr = load_pixel_shader("sdr/bezier.p.glsl"))) {
			return 1;
		}

		if(!(prog = create_program_link(vsdr, tcsdr, tesdr, psdr, 0))) {
			return 1;
		}
	}

	glutMainLoop();
	return 0;
}


void disp(void)
{
	float lpos[] = {-5, 10, 4, 1};

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0, 0, -cam_dist);
	glRotatef(cam_phi, 1, 0, 0);
	glRotatef(cam_theta, 0, 1, 0);

	glLightfv(GL_LIGHT0, GL_POSITION, lpos);

	set_material(0.2, 0.35, 1.0, 1.0, 1.0, 1.0, 60.0);

	draw_teapot(2.0);

	glutSwapBuffers();
}

void draw_teapot(float scale)
{
	int i;

	scale /= 2.0;

	glPushMatrix();
	glTranslatef(0, -3.15 * scale * 0.5, 0);
	glRotatef(-90.0, 1, 0, 0);
	glScalef(scale, scale, scale);

	glUseProgram(prog);
	set_uniform_int(prog, "tess_level", tess_level);
	set_uniform_float3(prog, "norm_scale", 1, 1, 1);

	glBegin(GL_PATCHES);

	/* first render the front-facing patches */
	for(i=0; i<NUM_TEAPOT_PATCHES; i++) {
		if(teapot_part_flip[i] > 0.0) {
			draw_teapot_patch(teapot_verts, teapot_index + i * 16, 0, teapot_part_rot[i]);
		}
	}
	glEnd();

	set_uniform_float3(prog, "norm_scale", -1, -1, -1);

	glFrontFace(GL_CW);
	glBegin(GL_PATCHES);
	/* then render the flipped ones */
	for(i=0; i<NUM_TEAPOT_PATCHES; i++) {
		if(teapot_part_flip[i] < 0.0) {
			draw_teapot_patch(teapot_verts, teapot_index + i * 16, 1, teapot_part_rot[i]);
		}
	}
	glEnd();
	glFrontFace(GL_CCW);

	glUseProgram(0);

	glPopMatrix();
}

void draw_teapot_patch(struct vec3 *bez_cp, int *index, int flip, float rot)
{
	int i;
	float cosr = cos(M_PI * rot / 180.0);
	float sinr = sin(M_PI * rot / 180.0);

	for(i=0; i<16; i++) {
		struct vec3 cp = bez_cp[index[i]];

		float x = cosr * cp.x + sinr * cp.y;
		float y = -sinr * cp.x + cosr * cp.y;

		glVertex3f(x, flip ? -y : y, cp.z);
	}
}

void set_material(float dr, float dg, float db, float sr, float sg, float sb, float shin)
{
	float dcol[4], scol[4];

	dcol[0] = dr;
	dcol[1] = dg;
	dcol[2] = db;
	dcol[3] = 1.0;

	scol[0] = sr;
	scol[1] = sg;
	scol[2] = sb;
	scol[3] = 1.0;

	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, dcol);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, scol);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, shin);
}

void reshape(int x, int y)
{
	glViewport(0, 0, x, y);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(50.0, (float)x / (float)y, 0.5, 500.0);
}

void keyb(unsigned char key, int x, int y)
{
	switch(key) {
	case 27:
		exit(0);

	case 'w':
		{
			static int wire;
			wire = !wire;
			if(wire) {
				glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
			} else {
				glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
			}
		}
		glutPostRedisplay();
		break;

	case '=':
		if(tess_level < max_tess_level) {
			tess_level++;
			printf("tessellation level: %d\n", tess_level);
			set_uniform_int(prog, "tess_level", tess_level);
			glutPostRedisplay();
		}
		break;

	case '-':
		if(tess_level > 1) {
			tess_level--;
			printf("tessellation level: %d\n", tess_level);
			set_uniform_int(prog, "tess_level", tess_level);
			glutPostRedisplay();
		}
		break;
	}
}

static int bnstate[16];
static int prev_x, prev_y;

void mouse(int bn, int state, int x, int y)
{
	int idx = bn - GLUT_LEFT_BUTTON;

	if(idx < sizeof bnstate / sizeof *bnstate) {
		bnstate[idx] = state == GLUT_DOWN;
	}
	prev_x = x;
	prev_y = y;
}

void motion(int x, int y)
{
	int dx = x - prev_x;
	int dy = y - prev_y;
	prev_x = x;
	prev_y = y;

	if(bnstate[0]) {
		cam_theta += dx * 0.5;
		cam_phi += dy * 0.5;
		if(cam_phi < -90) cam_phi = -90;
		if(cam_phi > 90) cam_phi = 90;
		glutPostRedisplay();
	}

	if(bnstate[2]) {
		cam_dist += dy * 0.1;
		glutPostRedisplay();
	}
}
